# -*- coding: utf-8 -*-

"""

WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

	src/wenv/_core/env.py: Managing a Wine-Python environment

	Copyright (C) 2017-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import json
from io import BytesIO
import os
import shutil
import subprocess
import sys
import zipfile

from .config import EnvConfig
from .const import c, COVERAGE_STARTUP, HELP_STR
from .source import download, PythonVersion

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# HELPER ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _symlink(src, dest):

	if not os.path.lexists(dest):
		os.symlink(src, dest)

	if not os.path.exists(dest):
		raise OSError('"{LINK:s}" could not be created'.format(LINK = dest))
	if not os.path.islink(dest):
		raise OSError('"{LINK:s}" is not a symlink'.format(LINK = dest))
	if os.readlink(dest) != src:
		raise OSError('"{LINK:s}" points to the wrong source'.format(LINK = dest))

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# PATHS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Paths:

	def __init__(self, pythonprefix, arch, pythonversion):

		self._pythonprefix = pythonprefix
		self._pythonversion_block = PythonVersion.from_config(arch, pythonversion).as_block()

	def __getitem__(self, key):

		if key == 'pythonprefix':
			return self._pythonprefix
		elif key == 'lib':
			return os.path.join(self['pythonprefix'], 'Lib')
		elif key == 'sitepackages':
			return os.path.join(self['lib'], 'site-packages')
		elif key == 'sitecustomize':
			return os.path.join(self['sitepackages'], 'sitecustomize.py')
		elif key == 'scripts':
			return os.path.join(self['pythonprefix'], 'Scripts')
		elif key == 'interpreter':
			return os.path.join(self['pythonprefix'], 'python.exe')
		elif key == 'pip':
			return os.path.join(self['scripts'], 'pip.exe')
		elif key == 'libzip':
			return os.path.join(self['pythonprefix'], 'python%s.zip' % self._pythonversion_block)
		elif key == 'pth':
			return os.path.join(self['pythonprefix'], 'python%s._pth' % self._pythonversion_block)
		else:
			raise KeyError('not a valid path key')

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# WINE-PYTHON ENVIRONMENT CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Env:

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# INIT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def __init__(self, **kwargs):

		if 'parameter' in kwargs.keys(): # legacy API
			if len(kwargs) > 1:
				raise ValueError('legacy API: only allows one parameter dict, named "parameter"')
			kwargs = kwargs['parameter']
			if kwargs is None:
				kwargs = dict()
			if not isinstance(kwargs, dict) and not isinstance(kwargs, EnvConfig):
				raise TypeError('legacy API: only allows one parameter, must be a dict or EnvConfig')
			if not isinstance(kwargs, EnvConfig):
				kwargs = EnvConfig(**kwargs)
		else:
			kwargs = EnvConfig(**kwargs)

		self._p = kwargs
		self._init_dicts()

	def _init_dicts(self):
		"""
		Initialize core dictionaries. Function can also be used for re-initialization.
		"""

		# Init Wine cmd names
		self._wine_dict = {'win32': 'wine', 'win64': 'wine64'}

		# Init Python environment paths
		self._path_dict = Paths(self._p['pythonprefix'], self._p['arch'], self._p['pythonversion'])
		# Init Python commands and scripts
		self._init_cmd_dict()
		# Init internal CLI commands
		self._init_cli_dict()
		# Init environment variables
		self._init_envvar_dict()

	def _init_cmd_dict(self):

		def ls_exe(dir):
			if not os.path.isdir(dir):
				return
			for item in os.listdir(dir):
				if not item.lower().endswith('.exe'):
					continue
				yield item[:-4], os.path.join(dir, item)

		self._cmd_dict = {
			item: path for item, path in ls_exe(self._path_dict['scripts'])
			}
		self._cmd_dict.update({
			item: path for item, path in ls_exe(self._path_dict['pythonprefix'])
			})

	def _init_cli_dict(self):

		self._cli_dict = {
			item[5:]: getattr(self, item)
			for item in dir(self)
			if item.startswith('_cli_') and hasattr(getattr(self, item), '__call__')
			}

	def _init_envvar_dict(self):

		self._envvar_dict = {k: os.environ[k] for k in os.environ.keys()} # HACK Required for Travis CI
		self._envvar_dict.update(dict(
			WINEARCH = self._p['arch'], # Architecture
			WINEPREFIX = self._p['wineprefix'], # Wine prefix / directory
			WINEDLLOVERRIDES = 'mscoree=d', # Disable MONO: https://unix.stackexchange.com/a/191609
			WINEDEBUG = self._p['winedebug'], # Wine debug level
			PYTHONHOME = self._p['pythonprefix'], # Python home for Wine Python (can be a Unix path)
			VIRTUAL_ENV = '', # Reset Unix virtual env variable - wenv is "independent"
			PIP_NO_WARN_SCRIPT_LOCATION = '0', # pip will not warn that pythonprefix and scripts are not in PATH
			))
		if self._p['wineinstallprefix'] not in (None, ''): # allow custom installations of Wine outside of PATH
			path = self._envvar_dict.get('PATH', '')
			self._envvar_dict['PATH'] = os.path.join(self._p['wineinstallprefix'], 'bin') + ':' + path
			ld_library_path = self._envvar_dict.get('LD_LIBRARY_PATH', '')
			self._envvar_dict['LD_LIBRARY_PATH'] = ':'.join((
				os.path.join(self._p['wineinstallprefix'], 'lib'),
				os.path.join(self._p['wineinstallprefix'], 'lib64'),
				ld_library_path
				)) # https://wiki.winehq.org/FAQ#Can_I_install_more_than_one_Wine_version_on_my_system.3F

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# WINE BUG #47766 / ZUGBRUECKE BUG #49
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def wine_47766_workaround(self):
		"""
		PathAllocCanonicalize treats path segments start with dots wrong.
		https://bugs.winehq.org/show_bug.cgi?id=47766
		"""

		is_clean = lambda path: not any([seg.startswith('.') for seg in path.split(os.path.sep)])

		pythonprefix = os.path.abspath(self._p['pythonprefix'])

		if pythonprefix != self._p['pythonprefix']:
			self._p['pythonprefix'] = pythonprefix
			self._init_dicts()

		if is_clean(self._p['pythonprefix']):
			return

		import tempfile, hashlib
		link_path = os.path.join(
			tempfile.gettempdir(),
			'wenv-' + hashlib.sha256(self._p['pythonprefix'].encode('utf-8')).hexdigest()[:8],
			)
		if not is_clean(link_path):
			raise OSError('unable to create clean link path: "{LINK:s}"'.format(LINK = link_path))

		if os.path.exists(self._p['pythonprefix']):
			_symlink(self._p['pythonprefix'], link_path)

		self._p['pythonprefix'] = link_path
		self._init_dicts()

	def wine_47766_workaround_uninstall(self):

		self.wine_47766_workaround()

		if os.path.lexists(self._p['pythonprefix']):
			os.unlink(self._p['pythonprefix'])

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ENSURE ENVIRONMENT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def ensure(self):

		self.setup_wineprefix()
		self.setup_pythonprefix()
		self.wine_47766_workaround() # must run after setup_pythonprefix and before setup_pip
		self.setup_pip()

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# DESTROY / UNINSTALL ENVIRONMENT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def uninstall(self):

		# Does Wine prefix exist?
		if os.path.exists(self._p['wineprefix']):
			# Delete tree
			shutil.rmtree(self._p['wineprefix'])

		# Does Python prefix exist?
		if os.path.exists(self._p['pythonprefix']):
			# Delete tree
			shutil.rmtree(self._p['pythonprefix'])

		self.wine_47766_workaround_uninstall()

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CACHE INSTALLATION FILES LOCALLY
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def cache(self):

		version = PythonVersion.from_config(self._p['arch'], self._p['pythonversion'])

		os.makedirs(self._p['cache'], exist_ok = True)

		with open(os.path.join(self._p['cache'], version.as_zipname()), 'wb') as f:
			f.write(self._get_python(offline = False))
		with open(os.path.join(self._p['cache'], 'get-pip.py'), 'wb') as f:
			f.write(self._get_pip(offline = False))

		for package in ('pip', 'setuptools', 'wheel'):
			self.cache_package(package)

	def cache_package(self, name):

		os.makedirs(self._p['packages'], exist_ok = True)

		meta = json.loads(
			download('https://pypi.org/pypi/%s/json' % name, mode = 'binary').decode('utf-8')
			)

		for item in meta['urls']:
			with open(os.path.join(self._p['packages'], item['filename']), 'wb') as f:
				f.write(download(item['url'], mode = 'binary'))

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Fetch installer data
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def _get_python(self, offline = False):

		version = PythonVersion.from_config(self._p['arch'], self._p['pythonversion'])

		if offline:
			with open(os.path.join(self._p['cache'], version.as_zipname()), 'rb') as f:
				return f.read()
		else:
			return download(version.as_url(), mode = 'binary')

	def _get_pip(self, offline = False):

		if offline:
			with open(os.path.join(self._p['cache'], 'get-pip.py'), 'rb') as f:
				return f.read()
		else:
			return download('https://bootstrap.pypa.io/get-pip.py', mode = 'text').encode('utf-8')

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def setup_wineprefix(self, overwrite = False):

		if not isinstance(overwrite, bool):
			raise TypeError('overwrite is not a boolean')

		# Does it exist?
		if os.path.exists(self._p['wineprefix']):
			# Exit if overwrite flag is not set
			if not overwrite:
				return
			# Delete if overwrite is set
			shutil.rmtree(self._p['wineprefix'])

		os.makedirs(self._p['wineprefix']) # HACK Wine on Travis CI expects folder to exist

		# Start wine server into prepared environment
		envvar_dict = self._envvar_dict.copy()
		envvar_dict['DISPLAY'] = ''
		proc = subprocess.Popen(
			['wine', 'wineboot', '-i'],
			env = envvar_dict
			)
		proc.wait()
		if proc.returncode != 0:
			sys.exit(1)

	def setup_pythonprefix(self, overwrite = False):

		if not isinstance(overwrite, bool):
			raise TypeError('overwrite is not a boolean')

		# Is there a pre-existing Python installation with identical parameters?
		preexisting = os.path.isfile(self._path_dict['interpreter'])

		# Is there a preexisting installation and should it be overwritten?
		if preexisting and overwrite:
			# Delete folder
			shutil.rmtree(self._p['pythonprefix'])

		# Make sure the target directory exists
		if not os.path.exists(self._p['pythonprefix']):
			# Create folder
			os.makedirs(self._p['pythonprefix'])

		# Only do if Python is not there OR if should be overwritten
		if overwrite or not preexisting:

			# Generate in-memory file-like-object
			archive_zip = BytesIO()
			# Fetch Python zip file
			archive_zip.write(self._get_python(self._p['offline']))
			# Unpack from memory to disk
			with zipfile.ZipFile(archive_zip) as f:
				f.extractall(path = self._p['pythonprefix']) # Directory created if required

			# Unpack Python library from embedded zip on disk
			with zipfile.ZipFile(self._path_dict['libzip'], 'r') as f:
				f.extractall(path = self._path_dict['lib']) # Directory created if required
			# Remove Python library zip from disk
			os.remove(self._path_dict['libzip'])

			# HACK: Fix library path in pth-file (CPython >= 3.6)
			with open(self._path_dict['pth'], 'w') as f:
				f.write('Lib\n.\n\n# Uncomment to run site.main() automatically\nimport site\n')

		# Create site-packages folder if it does not exist
		if not os.path.exists(self._path_dict['sitepackages']):
			# Create folder
			os.makedirs(self._path_dict['sitepackages'])

	def setup_pip(self):

		# Exit if it exists
		if os.path.isfile(self._path_dict['pip']):
			return

		envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
		envvar_dict.update(self._p.export_envvar_dict())

		if self._p['offline']:
			proc = subprocess.Popen([
				'wenv', 'python',
				os.path.join(self._p['cache'], 'get-pip.py'),
				'--no-index', '--find-links=%s' % self._p['packages'],
				], env = envvar_dict
				)
			proc.wait()
		else:
			getpip = self._get_pip(self._p['offline'])
			proc = subprocess.Popen(
				['wenv', 'python'],
				stdin = subprocess.PIPE,
				env = envvar_dict
				)
			proc.communicate(input = getpip)

	def setup_coverage_activate(self):

		# Ensure that coverage is started with the Python interpreter
		siteconfig_cnt = ''
		if os.path.isfile(self._path_dict['sitecustomize']):
			with open(self._path_dict['sitecustomize'], 'r') as f:
				siteconfig_cnt = f.read()
			if COVERAGE_STARTUP in siteconfig_cnt:
				return
		with open(self._path_dict['sitecustomize'], 'w') as f:
			f.write(siteconfig_cnt + '\n' + COVERAGE_STARTUP)

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# PACKAGE MANAGEMENT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def install_package(self, name, update = False):
		"""
		Thin wrapper for `wenv pip install`
		"""

		if not isinstance(name, str):
			raise TypeError('name must be str')
		if len(name) == 0:
			raise ValueError('name must not be empty')
		if not isinstance(update, bool):
			raise TypeError('update must be bool')

		cmd = ['wenv', 'pip', 'install']
		if update:
			cmd.append('-U')
		cmd.append(name)

		envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
		envvar_dict.update(self._p.export_envvar_dict())

		proc = subprocess.Popen(
			cmd,
			stdout = subprocess.PIPE, stderr = subprocess.PIPE,
			env = envvar_dict
			)
		outs, errs = proc.communicate()
		if proc.returncode != 0:
			raise SystemError('installing package "%s" failed' % name, outs, errs)

	def uninstall_package(self, name):
		"""
		Thin wrapper for `wenv pip uninstall -y`
		"""

		if not isinstance(name, str):
			raise TypeError('name must be str')
		if len(name) == 0:
			raise ValueError('name must not be empty')

		envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
		envvar_dict.update(self._p.export_envvar_dict())

		proc = subprocess.Popen(
			['wenv', 'pip', 'uninstall', '-y', name],
			stdout = subprocess.PIPE, stderr = subprocess.PIPE,
			env = envvar_dict
			)
		outs, errs = proc.communicate()
		if proc.returncode != 0:
			raise SystemError('uninstalling package "%s" failed' % name, outs, errs)

	def list_packages(self):
		"""
		Thin wrapper for `wenv pip list --format json`
		"""

		envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
		envvar_dict.update(self._p.export_envvar_dict())

		proc = subprocess.Popen(
			['wenv', 'pip', 'list', '--format', 'json'],
			stdout = subprocess.PIPE, stderr = subprocess.PIPE,
			env = envvar_dict
			)
		outs, errs = proc.communicate()
		if proc.returncode != 0:
			raise SystemError('uninstalling package "%s" failed' % name, outs, errs)

		return json.loads(outs.decode('utf-8'))

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLI
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def _cli_init(self):
		"sets up an environment (Wine prefix, Python interpreter, pip, setuptools, wheel)"

		self.ensure()

	def _cli_init_coverage(self):
		"enables coverage analysis inside wenv"

		self.setup_coverage_activate()

	def _cli_cache(self):
		"fetches installation files and caches them for offline usage (Python interpreter, pip, setuptools, wheel)"

		self.cache()

	def _cli_clean(self):
		"removes current environment (Python interpreter, pip, setuptools, wheel, all installed packages)"

		self.uninstall()

	def _cli_help(self):
		"prints this help text"

		def colorize(text):
			for color, key in (
				('GREEN', 'python'),
				('CYAN', 'pip'),
				('YELLOW', 'wheel'),
				('MAGENTA', 'pytest'),
				('MAGENTA', 'py.test'),
				):
				text = text.replace(key, c[color] + key + c['RESET'])
			return text

		sys.stdout.write(HELP_STR.format(
			CLIS = '\n'.join([
				'- wenv {CLI:s}: {HELP:s}'.format(
					CLI = key,
					HELP = self._cli_dict[key].__doc__,
					)
				for key in sorted(self._cli_dict.keys())
				]),
			SCRIPTS = colorize('\n'.join([
				'- wenv {SCRIPT:s}'.format(
					SCRIPT = key,
					)
				for key in sorted(self._cmd_dict.keys())
				]))
			))
		sys.stdout.flush()

	def cli(self):

		# No command passed
		if len(sys.argv) < 2:
			sys.stderr.write('There was no command passed.\n')
			sys.stderr.flush()
			sys.exit(1)

		# Separate command and arguments
		cmd, param = sys.argv[1], sys.argv[2:]

		# Allow -h and --help
		if cmd in ['-h', '--help']:
			cmd = 'help'

		# Special CLI command
		if cmd in self._cli_dict.keys():
			self._cli_dict[cmd]()
			sys.exit(0)

		# Command is unknown
		if cmd not in self._cmd_dict.keys():
			sys.stderr.write('Unknown command or script: "{CMD:s}"\n'.format(CMD = cmd))
			sys.stderr.flush()
			sys.exit(1)

		# Get Wine depending on arch
		wine = self._wine_dict[self._p['arch']]

		self.wine_47766_workaround()

		# Replace this process with Wine
		os.execvpe(
			wine,
			(wine, self._cmd_dict[cmd]) + tuple(param), # Python 3.4: No in-place unpacking of param
			self._envvar_dict
			)

	def shebang(self):
		"""Working around a lack of Unix specification ...
		https://stackoverflow.com/q/4303128/1672565
		https://unix.stackexchange.com/q/63979/28301
		https://lists.gnu.org/archive/html/bug-sh-utils/2002-04/msg00020.html
		"""

		if len(sys.argv) < 2:
			raise OSError('entry point meant to be used as a shebang but no file name was provided')

		# Get Wine depending on arch
		wine = self._wine_dict[self._p['arch']]

		self.wine_47766_workaround()

		# Replace this process with Wine
		os.execvpe(
			wine,
			(wine, self._cmd_dict['python'], sys.argv[1]),
			self._envvar_dict
			)

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLI EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def cli():

	Env().cli()

def shebang():

	Env().shebang()



# -*- coding: utf-8 -*-

"""

WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

	src/wenv/_core/source.py: Obtaining Python and pip locally or remotely

	Copyright (C) 2017-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from multiprocessing.pool import ThreadPool

import requests

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def download(down_url, mode = 'binary'):

	assert mode in ('text', 'binary')
	assert isinstance(down_url, str)

	r = requests.get(down_url)

	assert r.ok
	r.raise_for_status()

	if r.encoding is not None:
		assert mode == 'text' and isinstance(r.text, str)
		return r.text
	else:
		assert mode == 'binary' and isinstance(r.content, bytes)
		return r.content

def get_available_python_versions():

	versions = [
		tuple(int(nr) for nr in line.split('"')[1][:-1].split('.'))
		for line in download('https://www.python.org/ftp/python/', mode = 'text').split('\n')
		if all([
			line.startswith('<a href="'),
			line[9:10].isdigit(),
			not line.startswith('<a href="2.')
			])
		]
	version_urls = [
		'https://www.python.org/ftp/python/%d.%d.%d/' % version
			if len(version) == 3 else
		'https://www.python.org/ftp/python/%d.%d/' % version
		for version in versions
		]
	version_downloads = ThreadPool(8).imap(lambda x: download(x, mode = 'text'), version_urls)
	embedded_versions = {
		version: [
			python_version.from_zipname(line.split('"')[1])
			for line in version_download.split('\n')
			if all((line.startswith('<a href="'), 'embed' in line, '.zip' in line, '.zip.asc' not in line))
			]
		for version, version_url, version_download in zip(versions, version_urls, version_downloads)
		}
	embedded_versions = {k: v for k, v in embedded_versions.items() if len(v) > 0}

	sorted_versions = {
		'win32': {
			version_tuple: sorted([version for version in versions if version.arch == 'win32'])
			for version_tuple, versions in embedded_versions.items()
			},
		'win64': {
			version_tuple: sorted([version for version in versions if version.arch == 'win64'])
			for version_tuple, versions in embedded_versions.items()
			}
		}

	return sorted_versions

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: PYTHON VERSION
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class PythonVersion:

	def __init__(self, arch, major, minor, maintenance, build = 'stable'):

		if not isinstance(arch, str):
			raise TypeError('arch must be str')
		if arch not in ('win32', 'win64'):
			raise ValueError('Unknown arch: ' + arch)
		if any((not isinstance(item, int) for item in (major, minor, maintenance))):
			raise TypeError('Unknown type for major, minor or maintenance')
		if major <= 2:
			raise ValueError('Only Python 3 and newer supported')
		if not isinstance(build, str):
			raise TypeError('build must be str')

		self._arch = arch
		self._major, self._minor, self._maintenance = major, minor, maintenance
		self._build = 'stable' if build == '' else build

	def __str__(self):

		return '%d.%d.%d.%s' % (
			self._major, self._minor, self._maintenance, self._build
			)

	def __repr__(self):

		return '<Python %d.%d.%d.%s (%s)>' % (
			self._major, self._minor, self._maintenance, self._build, self._arch
			)

	def __eq__(self, other):
		return self._as_sort() == other._as_sort()
	def __gt__(self, other):
		return self._as_sort() > other._as_sort()
	def __lt__(self, other):
		return self._as_sort() < other._as_sort()
	def _as_sort(self):
		return '%04d-%04d-%04d-%s' % (self._major, self._minor, self._maintenance, self._build)

	@property
	def arch(self):

		return self._arch

	def as_block(self):

		return '%d%d' % (self._major, self._minor)

	def as_config(self):

		return str(self)

	def as_url(self):

		return 'https://www.python.org/ftp/python/%d.%d.%d/' % (
			self._major, self._minor, self._maintenance
			) + self.as_zipname()

	def as_zipname(self):

		build = '' if self._build == 'stable' else self._build
		arch = 'amd64' if self._arch == 'win64' else self._arch
		sub_tuple = (self._major, self._minor, self._maintenance, build, arch)

		if build.startswith('post'):
			return 'python-%d.%d.%d.%s-embed-%s.zip' % sub_tuple
		else:
			return 'python-%d.%d.%d%s-embed-%s.zip' % sub_tuple

	@classmethod
	def from_config(cls, arch, version):

		if not isinstance(version, str):
			raise TypeError('version must be str')
		segments = version.split('.')
		if not len(segments) in (3, 4):
			raise ValueError('wrong number of version segments')
		if len(segments) == 3:
			segments.append('stable')
		if not all((segment.isnumeric() for segment in segments[:3])):
			raise ValueError('version segments are not numeric')
		segments = tuple([int(segment) for segment in segments[:3]] + [segments[3]])

		return cls(arch, *segments)

	@classmethod
	def from_zipname(cls, zip_name):

		if not isinstance(zip_name, str):
			raise TypeError('zip_name must be str')

		fragments = zip_name.split('-')
		fragments.append(fragments[3].split('.')[1])
		fragments[3] = fragments[3].split('.')[0]

		if not fragments[0] == 'python':
			raise ValueError('fagment[0] != "python"')
		if not  fragments[2] == 'embed':
			raise ValueError('fagment[2] != "embed"')
		if not  fragments[3] in ('win32', 'amd64'):
			raise ValueError('fagment[3] not in in ("win32", "amd64")')
		if not  fragments[4] == 'zip':
			raise ValueError('fagment[4] != "zip"')

		arch = 'win32' if fragments[3] == 'win32' else 'win64'
		release = [
			int(f) if f.isnumeric() else f
			for f in fragments[1].split('.')
			]

		if isinstance(release[2], str):
			if not len(release[2]) > 0:
				raise ValueError('broken maintenance/build fragment')
			for pos, char in enumerate(release[2]):
				if not char.isdigit():
					break
			release.append(release[2][pos:])
			release[2] = release[2][:pos]
			if not release[2].isnumeric():
				raise ValueError('maintenance fragment not numeric')
			release[2] = int(release[2])

		if len(release) == 3:
			release.append('stable')
		if not len(release) == 4:
			raise ValueError('release does not have 4 fragments (major,minor,maintenance,build)')

		return cls(arch, *release)



# -*- coding: utf-8 -*-

"""

WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

	src/wenv/_core/config.py: Handles the modules configuration

	Copyright (C) 2017-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os
import json
import site
import sys

from .const import CONFIG_FN
from .errors import EnvConfigParserError

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONFIGURATION CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class EnvConfig(dict):

	def __init__(self, **override_dict):

		# Call parent constructur, just in case
		super().__init__()

		# Get config from files as a prioritized list
		for config in self._get_config_from_files():
			self.update(config)

		# Add override parameters
		if len(override_dict) > 0:
			self.update(override_dict)

	def __getitem__(self, key):

		env_var = 'WENV_{NAME:s}'.format(NAME = key.upper())
		if env_var in os.environ.keys():
			value = os.environ[env_var]
			if len(value) > 0:
				if value.isnumeric():
					return int(value)
				elif value.strip().lower() in ('true', 'false'):
					return {'true': True, 'false': False}[value.strip().lower()]
				else:
					return value

		if key in self.keys():
			return super().__getitem__(key)

		if key == 'arch':
			return 'win32' # Define Wine & Wine-Python architecture
		elif key == 'pythonversion':
			return '3.7.4' # Define Wine-Python version
		elif key == 'winedebug':
			return '-all' # Wine debug output off
		elif key == 'wineinstallprefix':
			return None # no custom Wine installation outside of PATH
		elif key == 'prefix':
			install_location = os.path.abspath(__file__)
			if install_location.startswith(site.USER_BASE): # Hacky way of looking for a user installation
				return site.USER_BASE
			else:
				return sys.prefix
		elif key == 'wineprefix':
			return os.path.join(self['prefix'], 'share', 'wenv', self['arch'])
		elif key == 'pythonprefix':
			return os.path.join(self['wineprefix'], 'drive_c', 'python-%s' % self['pythonversion'])
		elif key == 'offline':
			return False
		elif key == 'cache':
			return os.path.join(self['prefix'], 'share', 'wenv', 'cache')
		elif key == 'packages':
			return os.path.join(self['cache'], 'packages')
		elif key == '_issues_50_workaround':
			return False # Workaround for zugbruecke issue #50 (symlinks ...)
		else:
			raise KeyError('not a valid configuration key')

	def export_envvar_dict(self):

		return {
			'WENV_' + field.upper(): '' if field is None else str(self[field])
			for field in (
				'arch',
				'pythonversion',
				'winedebug',
				'wineinstallprefix',
				'prefix',
				'wineprefix',
				'pythonprefix',
				'offline',
				'cache',
				'packages',
				'_issues_50_workaround',
				)
			}

	def _get_config_from_files(self):

		# Look for config in the usual spots
		for fn in [
			'/etc/wenv',
			os.path.join('/etc', CONFIG_FN),
			os.path.join(os.path.expanduser('~'), CONFIG_FN),
			os.environ.get('WENV'),
			os.path.join(os.environ.get('WENV'), CONFIG_FN) if os.environ.get('WENV') is not None else None,
			os.path.join(os.getcwd(), CONFIG_FN),
			]:

			cnt_dict = self._load_config_from_file(fn)

			if cnt_dict is not None:
				yield cnt_dict

	def _load_config_from_file(self, try_path):

		# If there is a path ...
		if try_path is None:
			return

		# Is this a file?
		if not os.path.isfile(try_path):
			return

		# Read file
		try:
			with open(try_path, 'r', encoding = 'utf-8') as f:
				cnt = f.read()
		except:
			raise EnvConfigParserError('Config file could not be read: "%s"' % try_path)

		# Try to parse it
		try:
			cnt_dict = json.loads(cnt)
		except:
			raise EnvConfigParserError('Config file could not be parsed: "%s"' % try_path)

		# Ensure that config has the right format
		if not isinstance(cnt_dict, dict):
			raise EnvConfigParserError('Config file is malformed: "%s"' % try_path)

		return cnt_dict



# -*- coding: utf-8 -*-

"""

WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

	setup.py: Used for package distribution

	Copyright (C) 2017-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from setuptools import (
	find_packages,
	setup
	)
import os
from sys import platform, version_info


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


# Bump version HERE!
_version_ = '0.2.1'


# List all versions of Python which are supported
python_minor_min = 4
python_minor_max = 8
confirmed_python_versions = [
	'Programming Language :: Python :: 3.{MINOR:d}'.format(MINOR = minor)
	for minor in range(python_minor_min, python_minor_max + 1)
	]


# Fetch readme file
with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
	long_description = f.read()


# Just in case someone is actually running this on Windows ...
if platform.startswith('win'):
	raise SystemExit('You are already running Windows. No need for this package!')


# Python 3.4 dependency / CI fix
pls = 'python-language-server'
assert version_info.major == 3
if version_info.minor <= 4:
	pls += '<0.32.0'


setup(
	name = 'wenv',
	packages = find_packages('src'),
	package_dir = {'': 'src'},
	version = _version_,
	description = 'Running Python on Wine',
	long_description = long_description,
	long_description_content_type = 'text/markdown',
	author = 'Sebastian M. Ernst',
	author_email = 'ernst@pleiszenburg.de',
	url = 'https://github.com/pleiszenburg/wenv',
	download_url = 'https://github.com/pleiszenburg/wenv/archive/v%s.tar.gz' % _version_,
	license = 'LGPLv2',
	keywords = ['wine', 'cross platform'],
	scripts = [],
	include_package_data = True,
	python_requires = '>=3.{MINOR:d}'.format(MINOR = python_minor_min),
	install_requires = [
		'requests'
		],
	extras_require = {
		'dev': [
			'pytest',
			'coverage',
			'pytest-cov',
			pls,
			'setuptools',
			'Sphinx',
			'sphinx_rtd_theme',
			'twine',
			'wheel'
			]
		},
	zip_safe = False,
	entry_points = {'console_scripts': [
		'wenv = wenv:cli',
		'_wenv_python = wenv:shebang'
		]},
	classifiers = [
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Intended Audience :: Information Technology',
		'Intended Audience :: Science/Research',
		'Intended Audience :: System Administrators',
		'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
		'Operating System :: MacOS',
		'Operating System :: POSIX :: BSD',
		'Operating System :: POSIX :: Linux',
		'Programming Language :: Python :: 3'
		] + confirmed_python_versions + [
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: Implementation :: CPython',
		'Topic :: Scientific/Engineering',
		'Topic :: Software Development',
		'Topic :: System :: Operating System',
		'Topic :: System :: Operating System Kernels',
		'Topic :: Utilities'
		]
	)
